/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

#include "CbmKfTrackFitter.h"

#include "CaDefs.h"
#include "CaFramework.h"
#include "CaSimd.h"
#include "CaStation.h"
#include "CaTrackFit.h"
#include "CbmGlobalTrack.h"
#include "CbmL1.h"
#include "CbmL1Util.h"
#include "CbmMuchPixelHit.h"
#include "CbmMuchTrack.h"
#include "CbmMuchTrackingInterface.h"
#include "CbmMvdHit.h"
#include "CbmMvdTrackingInterface.h"
#include "CbmStsAddress.h"
#include "CbmStsHit.h"
#include "CbmStsSetup.h"
#include "CbmStsTrack.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTofHit.h"
#include "CbmTofTrack.h"
#include "CbmTofTrackingInterface.h"
#include "CbmTrdHit.h"
#include "CbmTrdTrack.h"
#include "CbmTrdTrackingInterface.h"
#include "FairRootManager.h"
#include "KFParticleDatabase.h"
#include "KfTrackParam.h"
#include "TClonesArray.h"
#include "TDatabasePDG.h"

using std::vector;
using namespace std;
using ca::fmask;
using ca::fvec;

namespace
{
  using namespace cbm::algo;
}

void CbmKfTrackFitter::Track::OrderNodesInZ()
{
  // sort the nodes in z
  std::sort(fNodes.begin(), fNodes.end(), [](const FitNode& a, const FitNode& b) { return a.fZ < b.fZ; });
}


CbmKfTrackFitter::CbmKfTrackFitter() {}

CbmKfTrackFitter::~CbmKfTrackFitter() {}

void CbmKfTrackFitter::Init()
{
  if (fIsInitialized) {
    return;
  }

  if (!CbmL1::Instance() || !CbmL1::Instance()->fpAlgo) {
    LOG(fatal) << "CbmKfTrackFitter: no CbmL1 task initialized ";
  }

  FairRootManager* ioman = FairRootManager::Instance();

  if (!ioman) {
    LOG(fatal) << "CbmKfTrackFitter: no FairRootManager";
  }

  // Get hits

  fInputMvdHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("MvdHit"));
  fInputStsHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("StsHit"));
  fInputTrdHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("TrdHit"));
  fInputMuchHits = dynamic_cast<TClonesArray*>(ioman->GetObject("MuchHit"));
  fInputTofHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("TofHit"));

  // Get global tracks
  fInputGlobalTracks = dynamic_cast<TClonesArray*>(ioman->GetObject("GlobalTrack"));

  // Get detector tracks
  fInputStsTracks  = dynamic_cast<TClonesArray*>(ioman->GetObject("StsTrack"));
  fInputMuchTracks = dynamic_cast<TClonesArray*>(ioman->GetObject("MuchTrack"));
  fInputTrdTracks  = dynamic_cast<TClonesArray*>(ioman->GetObject("TrdTrack"));
  fInputTofTracks  = dynamic_cast<TClonesArray*>(ioman->GetObject("TofTrack"));

  fIsInitialized = true;
}

void CbmKfTrackFitter::SetParticleHypothesis(int pdg)
{
  TParticlePDG* particlePDG = TDatabasePDG::Instance()->GetParticle(pdg);
  if (!particlePDG) {
    LOG(fatal) << "CbmKfTrackFitter: particle PDG " << pdg << " is not in the data base, please set the mass manually";
    return;
  }
  fMass       = particlePDG->Mass();
  fIsElectron = (abs(pdg) == 11);
}

void CbmKfTrackFitter::SetMassHypothesis(double mass)
{
  assert(mass >= 0.);
  fMass = mass;
}


void CbmKfTrackFitter::SetElectronFlag(bool isElectron) { fIsElectron = isElectron; }


bool CbmKfTrackFitter::CreateMvdStsTrack(CbmKfTrackFitter::Track& kfTrack, int stsTrackIndex)
{
  Init();
  if (!fIsInitialized) {
    return false;
  }

  if (!fInputStsTracks) {
    LOG(error) << "CbmKfTrackFitter: Sts track array not found!";
    return false;
  }
  if (stsTrackIndex >= fInputStsTracks->GetEntriesFast()) {
    LOG(error) << "CbmKfTrackFitter: Sts track index " << stsTrackIndex << " is out of range!";
    return false;
  }
  auto* stsTrack = dynamic_cast<const CbmStsTrack*>(fInputStsTracks->At(stsTrackIndex));
  if (!stsTrack) {
    LOG(error) << "CbmKfTrackFitter: Sts track is null!";
    return false;
  }

  CbmGlobalTrack globalTrack;
  globalTrack.SetStsTrackIndex(stsTrackIndex);
  globalTrack.SetParamFirst(*dynamic_cast<const FairTrackParam*>(stsTrack->GetParamFirst()));

  return CreateGlobalTrack(kfTrack, globalTrack);
}


bool CbmKfTrackFitter::CreateGlobalTrack(CbmKfTrackFitter::Track& kfTrack, int globalTrackIndex)
{
  Init();
  if (!fIsInitialized) {
    return false;
  }

  if (!fInputGlobalTracks) {
    LOG(error) << "CbmKfTrackFitter: Global track array not found!";
    return false;
  }

  if (globalTrackIndex >= fInputGlobalTracks->GetEntriesFast()) {
    LOG(error) << "CbmKfTrackFitter: Global track index " << globalTrackIndex << " is out of range!";
    return false;
  }

  auto* globalTrack = dynamic_cast<const CbmGlobalTrack*>(fInputGlobalTracks->At(globalTrackIndex));
  if (!globalTrack) {
    LOG(error) << "CbmKfTrackFitter: Global track is null!";
    return false;
  }

  return CreateGlobalTrack(kfTrack, *globalTrack);
}


bool CbmKfTrackFitter::CreateGlobalTrack(CbmKfTrackFitter::Track& kfTrack, const CbmGlobalTrack& globalTrack)
{
  kfTrack = {};
  Init();
  if (!fIsInitialized) {
    return false;
  }

  const ca::Parameters<ca::fvec>& caPar = CbmL1::Instance()->fpAlgo->GetParameters();

  int nStations = caPar.GetNstationsActive();

  CbmKfTrackFitter::Track t;

  {
    t.fNodes.resize(nStations);
    for (int i = 0; i < nStations; i++) {
      CbmKfTrackFitter::FitNode& n = t.fNodes[i];
      n                            = {};
      n.fMaterialLayer             = i;
      n.fZ                         = caPar.GetStation(i).GetZ<float>();
      n.fRadThick                  = 0.;
      n.fIsRadThickFixed           = false;
      n.fIsFitted                  = false;
      n.fIsXySet                   = false;
      n.fIsTimeSet                 = false;
      n.fHitSystemId               = ECbmModuleId::kNotExist;
      n.fHitAddress                = -1;
      n.fHitIndex                  = -1;
    }
  }

  // lambda to set the node from the pixel hit
  auto setNode = [&](const CbmPixelHit& h, int stIdx, int hitIdx, ca::EDetectorID detId) -> CbmKfTrackFitter::FitNode* {
    int ista = caPar.GetStationIndexActive(stIdx, detId);
    if (ista < 0) {
      return nullptr;
    }
    assert(ista < nStations);
    CbmKfTrackFitter::FitNode& n = t.fNodes[ista];
    n.fZ                         = h.GetZ();
    n.fMxy.SetX(h.GetX());
    n.fMxy.SetY(h.GetY());
    n.fMxy.SetDx2(h.GetDx() * h.GetDx());
    n.fMxy.SetDy2(h.GetDy() * h.GetDy());
    n.fMxy.SetDxy(h.GetDxy());
    n.fMxy.SetNdfX(1);
    n.fMxy.SetNdfY(1);
    n.fIsXySet = true;
    n.fMt.SetT(h.GetTime());
    n.fMt.SetDt2(h.GetTimeError() * h.GetTimeError());
    n.fMt.SetNdfT(1);
    n.fIsTimeSet       = (detId != ca::EDetectorID::kMvd);
    n.fRadThick        = 0.;
    n.fIsRadThickFixed = false;
    n.fHitSystemId     = cbm::L1Util::ConvertDetSystemId(detId);
    n.fHitAddress      = h.GetAddress();
    n.fHitIndex        = hitIdx;
    return &n;
  };

  // Read MVD & STS hits

  if (globalTrack.GetStsTrackIndex() >= 0) {

    int stsTrackIndex = globalTrack.GetStsTrackIndex();

    if (!fInputStsTracks) {
      LOG(error) << "CbmKfTrackFitter: Sts track array not found!";
      return false;
    }
    if (stsTrackIndex >= fInputStsTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Sts track index " << stsTrackIndex << " is out of range!";
      return false;
    }
    auto* stsTrack = dynamic_cast<const CbmStsTrack*>(fInputStsTracks->At(stsTrackIndex));
    if (!stsTrack) {
      LOG(error) << "CbmKfTrackFitter: Sts track is null!";
      return false;
    }

    // Read MVD hits

    int nMvdHits = stsTrack->GetNofMvdHits();
    if (nMvdHits > 0) {
      if (!fInputMvdHits) {
        LOG(error) << "CbmKfTrackFitter: Mvd hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nMvdHits; ih++) {
        int hitIndex = stsTrack->GetMvdHitIndex(ih);
        if (hitIndex >= fInputMvdHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Mvd hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmMvdHit*>(fInputMvdHits->At(hitIndex));
        setNode(hit, CbmMvdTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kMvd);
      }
    }

    // Read STS hits

    int nStsHits = stsTrack->GetNofStsHits();
    if (nStsHits > 0) {
      if (!fInputStsHits) {
        LOG(error) << "CbmKfTrackFitter: Sts hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nStsHits; ih++) {
        int hitIndex = stsTrack->GetStsHitIndex(ih);
        if (hitIndex >= fInputStsHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Sts hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmStsHit*>(fInputStsHits->At(hitIndex));
        setNode(hit, CbmStsTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kSts);
      }
    }
  }  // MVD & STS hits


  // Read Much hits

  if (globalTrack.GetMuchTrackIndex() >= 0) {
    int muchTrackIndex = globalTrack.GetMuchTrackIndex();
    if (!fInputMuchTracks) {
      LOG(error) << "CbmKfTrackFitter: Much track array not found!";
      return false;
    }
    if (muchTrackIndex >= fInputMuchTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Much track index " << muchTrackIndex << " is out of range!";
      return false;
    }
    auto* track = dynamic_cast<const CbmMuchTrack*>(fInputMuchTracks->At(muchTrackIndex));
    if (!track) {
      LOG(error) << "CbmKfTrackFitter: Much track is null!";
      return false;
    }
    int nHits = track->GetNofHits();
    if (nHits > 0) {
      if (!fInputMuchHits) {
        LOG(error) << "CbmKfTrackFitter: Much hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nHits; ih++) {
        int hitIndex = track->GetHitIndex(ih);
        if (hitIndex >= fInputMuchHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Much hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmMuchPixelHit*>(fInputMuchHits->At(hitIndex));
        setNode(hit, CbmMuchTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kMuch);
      }
    }
  }

  // Read TRD hits

  if (globalTrack.GetTrdTrackIndex() >= 0) {
    int trdTrackIndex = globalTrack.GetTrdTrackIndex();
    if (!fInputTrdTracks) {
      LOG(error) << "CbmKfTrackFitter: Trd track array not found!";
      return false;
    }
    if (trdTrackIndex >= fInputTrdTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Trd track index " << trdTrackIndex << " is out of range!";
      return false;
    }
    auto* track = dynamic_cast<const CbmTrdTrack*>(fInputTrdTracks->At(trdTrackIndex));
    if (!track) {
      LOG(error) << "CbmKfTrackFitter: Trd track is null!";
      return false;
    }
    int nHits = track->GetNofHits();
    if (nHits > 0) {
      if (!fInputTrdHits) {
        LOG(error) << "CbmKfTrackFitter: Trd hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nHits; ih++) {
        int hitIndex = track->GetHitIndex(ih);
        if (hitIndex >= fInputTrdHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Trd hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmTrdHit*>(fInputTrdHits->At(hitIndex));

        auto* node = setNode(hit, CbmTrdTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                             ca::EDetectorID::kTrd);
        if (node != nullptr && hit.GetClassType() == 1) {
          node->fHitSystemId = ECbmModuleId::kTrd2d;
        }
      }
    }
  }


  // Read TOF hits

  if (globalTrack.GetTofTrackIndex() >= 0) {
    int tofTrackIndex = globalTrack.GetTofTrackIndex();
    if (!fInputTofTracks) {
      LOG(error) << "CbmKfTrackFitter: Trd track array not found!";
      return false;
    }
    if (tofTrackIndex >= fInputTofTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Trd track index " << tofTrackIndex << " is out of range!";
      return false;
    }
    auto* track = dynamic_cast<const CbmTofTrack*>(fInputTofTracks->At(tofTrackIndex));
    if (!track) {
      LOG(error) << "CbmKfTrackFitter: Tof track is null!";
      return false;
    }

    int nHits = track->GetNofHits();
    if (nHits > 0) {
      if (!fInputTofHits) {
        LOG(error) << "CbmKfTrackFitter: Tof hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nHits; ih++) {
        int hitIndex = track->GetHitIndex(ih);
        if (hitIndex >= fInputTofHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Tof hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmTofHit*>(fInputTofHits->At(hitIndex));
        setNode(hit, CbmTofTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kTof);
      }
    }
  }

  t.OrderNodesInZ();

  // set initial track parameters
  // kf::TrackParamD tmp = cbm::L1Util::ConvertTrackParam(*globalTrack.GetParamFirst());
  // t.fNodes[0].fTrack.Set(tmp);
  // t.fNodes[0].fIsFitted = 1;

  kfTrack = t;
  return true;

  //  return CreateTrack(kfTrack, *globalTrack.GetParamFirst(), mvdHits, stsHits, muchHits, trdHits, tofHits);
}


void CbmKfTrackFitter::FilterFirstMeasurement(const FitNode& n)
{
  // a special routine to filter the first measurement.
  // the measurement errors are simply copied to the track covariance matrix

  assert(n.fIsXySet);

  const auto& mxy = n.fMxy;
  const auto& mt  = n.fMt;

  auto& tr = fFit.Tr();

  tr.ResetErrors(mxy.Dx2(), mxy.Dy2(), 100., 100., 1., 1.e4, 1.e2);
  tr.SetC10(mxy.Dxy());

  if (!(fSkipUnmeasuredCoordinates && mxy.NdfX() == 0)) {
    tr.SetX(mxy.X());
  }
  if (!(fSkipUnmeasuredCoordinates && mxy.NdfY() == 0)) {
    tr.SetY(mxy.Y());
  }
  tr.SetChiSq(0.);
  tr.SetChiSqTime(0.);
  tr.SetNdf(-5 + mxy.NdfX() + mxy.NdfY());
  if (n.fIsTimeSet) {
    tr.SetTime(mt.T());
    tr.SetC55(mt.Dt2());
    tr.SetNdfTime(-2 + 1);
  }
  else {
    tr.SetNdfTime(-2);
  }
  tr.Vi() = constants::phys::SpeedOfLightInv;
  tr.InitVelocityRange(0.5);
}


void CbmKfTrackFitter::AddMaterialEffects(CbmKfTrackFitter::FitNode& n, ca::FitDirection direction)
{
  // add material effects
  if (n.fMaterialLayer < 0) {
    return;
  }

  auto& tr = n.fIsFitted ? n.fTrack : fFit.Tr();

  // calculate the radiation thickness from the current track
  if (!n.fIsRadThickFixed) {
    n.fRadThick =
      CbmL1::Instance()->fpAlgo->GetParameters().GetMaterialThicknessScal(n.fMaterialLayer, tr.GetX(), tr.GetY());
  }

  double msQp0 = fIsQpForMsFixed ? fDefaultQpForMs : fFit.Qp0();

  fFit.MultipleScattering(n.fRadThick, msQp0);
  if (!fIsQpForMsFixed) {
    fFit.EnergyLossCorrection(n.fRadThick, direction);
  }
}


bool CbmKfTrackFitter::FitTrack(CbmKfTrackFitter::Track& t)
{
  // fit the track
  if (fVerbosityLevel > 0) {
    std::cout << "FitTrack ... " << std::endl;
  }
  bool ok = true;

  // ensure that the fitter is initialized
  Init();

  int nNodes = t.fNodes.size();

  if (nNodes <= 0) {
    LOG(warning) << "CbmKfTrackFitter: no nodes found!";
    return false;
  }

  int firstHitNode = -1;
  int lastHitNode  = -1;
  bool isOrdered   = true;

  {  // find the first and the last hit nodes. Check if the nodes are ordered in Z
    double zOld = -1.e10;
    for (int i = 0; i < nNodes; i++) {
      if (t.fNodes[i].fIsXySet) {
        if (firstHitNode < 0) {
          firstHitNode = i;
        }
        lastHitNode = i;
      }
      if (t.fNodes[i].fZ < zOld) {
        isOrdered = false;
      }
      zOld = t.fNodes[i].fZ;
    }
  }

  if (firstHitNode < 0 || lastHitNode < 0) {
    LOG(warning) << "CbmKfTrackFitter: no hit nodes found!";
    return false;
  }

  if (!isOrdered) {
    LOG(warning) << "CbmKfTrackFitter: the nodes are not ordered in Z!";
  }

  fFit.SetParticleMass(fMass);

  ca::FieldRegion<double> field;
  field.SetUseOriginalField();

  {  // fit downstream

    {  // initialisation at the first measurement
      FitNode& n = t.fNodes[firstHitNode];
      assert(n.fIsXySet);

      if (n.fIsFitted) {  // The initial parameters are taken from the previous fit
                          // Linearisation at the previously fitted momentum
        assert(n.fZ == n.fTrack.GetZ());
        fFit.SetTrack(n.fTrack);
      }
      else {  // The initial parameters are calculated from the first and the last measurements
        auto& tr  = fFit.Tr();
        auto& n1  = t.fNodes[lastHitNode];
        tr.X()    = n.fMxy.X();
        tr.Y()    = n.fMxy.Y();
        tr.Z()    = n.fZ;
        double dz = n1.fZ - n.fZ;
        tr.Tx()   = (n1.fMxy.X() - n.fMxy.X()) / dz;
        tr.Ty()   = (n1.fMxy.Y() - n.fMxy.Y()) / dz;
        tr.Qp()   = 0.;
        tr.Time() = 0.;
        tr.Vi()   = constants::phys::SpeedOfLightInv;
        fFit.SetQp0(0.);  // Linearisation at the infinite momentum
      }

      FilterFirstMeasurement(n);

      if (fVerbosityLevel > 0) {
        std::cout << " fit downstream: node " << firstHitNode << " chi2 " << fFit.Tr().GetChiSq() << " x "
                  << fFit.Tr().GetX() << " y " << fFit.Tr().GetY() << " tx " << fFit.Tr().GetTx() << " ty "
                  << fFit.Tr().GetTy() << std::endl;
      }
    }

    for (int iNode = firstHitNode + 1; iNode < nNodes; iNode++) {

      FitNode& n = t.fNodes[iNode];

      fFit.Extrapolate(n.fZ, field);

      if (fDoSmooth) {  // store the partially fitted track
        n.fTrack    = fFit.Tr();
        n.fIsFitted = false;  // the stored track is not fully fitted
      }

      if (n.fIsFitted && (iNode <= lastHitNode)) {  // linearisation is taken from the previous fit
        fFit.SetQp0(n.fTrack.GetQp());
      }

      AddMaterialEffects(n, FitDirection::kDownstream);

      if (n.fIsXySet) {
        fFit.FilterXY(n.fMxy, fSkipUnmeasuredCoordinates);
      }
      if (n.fIsTimeSet) {
        fFit.FilterTime(n.fMt);
      }

      if (fVerbosityLevel > 0) {
        std::cout << " fit downstream: node " << iNode << " chi2 " << fFit.Tr().GetChiSq() << " x " << fFit.Tr().GetX()
                  << " y " << fFit.Tr().GetY() << " tx " << fFit.Tr().GetTx() << " ty " << fFit.Tr().GetTy()
                  << std::endl;
      }
      if (iNode >= lastHitNode) {  // the track is fully fitted, store it
        n.fTrack    = fFit.Tr();
        n.fIsFitted = true;
        fFit.SetQp0(fFit.Tr().GetQp());  // set the linearisation of q/p to the fitted value
      }
    }
  }

  double dnChi2 = fFit.Tr().GetChiSq();
  double dnNdf  = fFit.Tr().GetNdf();

  {  // fit upstream

    {
      FitNode& n = t.fNodes[lastHitNode];
      assert(n.fIsFitted);
      fFit.SetTrack(n.fTrack);
      FilterFirstMeasurement(n);
      if (fVerbosityLevel > 0) {
        std::cout << "\n\n up node " << lastHitNode << " chi2 " << fFit.Tr().GetChiSq() << " x " << fFit.Tr().GetX()
                  << " y " << fFit.Tr().GetY() << " tx " << fFit.Tr().GetTx() << " ty " << fFit.Tr().GetTy()
                  << std::endl;
      }
    }

    for (int iNode = lastHitNode - 1; iNode >= 0; iNode--) {
      FitNode& n = t.fNodes[iNode];

      fFit.Extrapolate(n.fZ, field);

      AddMaterialEffects(n, FitDirection::kUpstream);

      if (n.fIsXySet) {
        fFit.FilterXY(n.fMxy, fSkipUnmeasuredCoordinates);
      }
      if (n.fIsTimeSet) {
        fFit.FilterTime(n.fMt);
      }
      if (fVerbosityLevel > 0) {
        std::cout << " up node " << iNode << " chi2 " << fFit.Tr().GetChiSq() << " x " << fFit.Tr().GetX() << " y "
                  << fFit.Tr().GetY() << " tx " << fFit.Tr().GetTx() << " ty " << fFit.Tr().GetTy() << std::endl;
      }

      // combine partially fitted downstream and upstream tracks
      n.fIsFitted = true;

      if (iNode > firstHitNode) {
        n.fIsFitted = false;
        if (fDoSmooth) {
          n.fIsFitted = Smooth(n.fTrack, fFit.Tr());
        }
      }
      else {
        n.fIsFitted = true;
        n.fTrack    = fFit.Tr();
      }
      if (fDoSmooth && !n.fIsFitted) {
        ok = false;
      }
      if (n.fIsFitted) {
        fFit.SetQp0(n.fTrack.GetQp());
      }
      AddMaterialEffects(n, FitDirection::kUpstream);
    }
  }

  double upChi2 = fFit.Tr().GetChiSq();
  double upNdf  = fFit.Tr().GetNdf();

  if (!fDoSmooth) {
    if (upNdf != dnNdf) {
      LOG(fatal) << "CbmKfTrackFitter: ndf mismatch: dn " << dnNdf << " != up " << upNdf;
    }
    if (fabs(upChi2 - dnChi2) > 1.e-2) {
      LOG(fatal) << "CbmKfTrackFitter: chi2 mismatch: dn " << dnChi2 << " != up " << upChi2 << " first node "
                 << firstHitNode << " last node " << lastHitNode << " of " << nNodes;
    }
  }
  // distribute the final chi2, ndf to all nodes

  const auto& tt = fFit.Tr();
  for (int iNode = 0; iNode < nNodes; iNode++) {
    FitNode& n = t.fNodes[iNode];
    if (iNode == lastHitNode) {
      //continue;
    }
    if (!fDoSmooth && iNode != lastHitNode) {
      //n.fIsFitted = false;
    }
    n.fTrack.SetNdf(tt.GetNdf());
    n.fTrack.SetNdfTime(tt.GetNdfTime());
    n.fTrack.SetChiSq(tt.GetChiSq());
    n.fTrack.SetChiSqTime(tt.GetChiSqTime());
  }

  return ok;
}

bool CbmKfTrackFitter::Smooth(kf::TrackParamBase<double>& t1, const kf::TrackParamBase<double>& t2)
{
  // TODO: move to the CaTrackFit class

  constexpr int nPar = kf::TrackParamV::kNtrackParam;
  constexpr int nCov = kf::TrackParamV::kNcovParam;

  double r[nPar] = {t1.X(), t1.Y(), t1.Tx(), t1.Ty(), t1.Qp(), t1.Time(), t1.Vi()};
  double m[nPar] = {t2.X(), t2.Y(), t2.Tx(), t2.Ty(), t2.Qp(), t2.Time(), t2.Vi()};

  double S[nCov], S1[nCov], Tmp[nCov];
  for (int i = 0; i < nCov; i++) {
    S[i] = (t1.GetCovMatrix()[i] + t2.GetCovMatrix()[i]);
  }

  int nullty = 0;
  int ifault = 0;
  cbm::algo::kf::utils::math::SymInv(S, nPar, S1, Tmp, &nullty, &ifault);

  if ((0 != ifault) || (0 != nullty)) {
    return false;
  }

  double dzeta[nPar];

  for (int i = 0; i < nPar; i++) {
    dzeta[i] = m[i] - r[i];
  }

  double C[nPar][nPar];
  double Si[nPar][nPar];
  for (int i = 0, k = 0; i < nPar; i++) {
    for (int j = 0; j <= i; j++, k++) {
      C[i][j]  = t1.GetCovMatrix()[k];
      Si[i][j] = S1[k];
      C[j][i]  = C[i][j];
      Si[j][i] = Si[i][j];
    }
  }

  // K = C * S^{-1}
  double K[nPar][nPar];
  for (int i = 0; i < nPar; i++) {
    for (int j = 0; j < nPar; j++) {
      K[i][j] = 0.;
      for (int k = 0; k < nPar; k++) {
        K[i][j] += C[i][k] * Si[k][j];
      }
    }
  }

  for (int i = 0, k = 0; i < nPar; i++) {
    for (int j = 0; j <= i; j++, k++) {
      double kc = 0.;  // K*C[i][j]
      for (int l = 0; l < nPar; l++) {
        kc += K[i][l] * C[l][j];
      }
      t1.CovMatrix()[k] = t1.CovMatrix()[k] - kc;
    }
  }

  for (int i = 0; i < nPar; i++) {
    double kd = 0.;
    for (int j = 0; j < nPar; j++) {
      kd += K[i][j] * dzeta[j];
    }
    r[i] += kd;
  }
  t1.X()    = r[0];
  t1.Y()    = r[1];
  t1.Tx()   = r[2];
  t1.Ty()   = r[3];
  t1.Qp()   = r[4];
  t1.Time() = r[5];
  t1.Vi()   = r[6];

  double chi2     = 0.;
  double chi2Time = 0.;
  for (int i = 0; i < nPar; i++) {
    double SiDzeta = 0.;
    for (int j = 0; j < nPar; j++) {
      SiDzeta += Si[i][j] * dzeta[j];
    }
    if (i < 5) {
      chi2 += dzeta[i] * SiDzeta;
    }
    else {
      chi2Time += dzeta[i] * SiDzeta;
    }
  }
  t1.SetChiSq(chi2 + t1.GetChiSq() + t2.GetChiSq());
  t1.SetChiSqTime(chi2Time + t1.GetChiSqTime() + t2.GetChiSqTime());
  t1.SetNdf(5 + t1.GetNdf() + t2.GetNdf());
  t1.SetNdfTime(2 + t1.GetNdfTime() + t2.GetNdfTime());
  return true;
}
